package vavetam.implementation.fileio;

import static org.junit.Assert.assertEquals;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import vavetam.api.structure.IDescriptor;
import vavetam.api.structure.IPredicate;
import vavetam.api.structure.IReferentialObject;
import vavetam.implementation.generation.incremental.formulation.Dictionary.Gender;
import vavetam.implementation.structure.Descriptor;
import vavetam.implementation.structure.Predicate;
import vavetam.implementation.structure.enums.Definiteness;
import vavetam.implementation.structure.enums.PredicateName;
import vavetam.implementation.structure.enums.RefOType;
import vavetam.implementation.structure.refo.ReferentialObjectBuilder;


public class XMLParser {
	
	private Logger log = Logger.getLogger(XMLParser.class);


	public Map<String, IReferentialObject> parse(File xmlfile)
			throws SAXException, IOException, ParserConfigurationException {
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		DocumentBuilder db = dbf.newDocumentBuilder();
		log.info("using " + xmlfile.getName() + " as map of referential objects");
		log.debug("starting to parse " + xmlfile.getName());
		Document doc = db.parse(xmlfile);
		Element mapElement = doc.getDocumentElement();
		assertEquals("top level element not called \"vavetam:map\"", "vavetam:map", mapElement.getNodeName());
		Map<String, ReferentialObjectBuilder> refOs = new HashMap<String,ReferentialObjectBuilder>();
		NodeList nl = mapElement.getChildNodes();
		log.debug("starting to create referential objects");
		for (int i = 0; i < nl.getLength(); i++) {
			Node refOnode = nl.item(i);
			if (!refOnode.getNodeName().equals("vavetam:referential-object")) {
				continue;
			}
			NamedNodeMap attributes = refOnode.getAttributes();
			String identifier = attributes.getNamedItem("id").getNodeValue();
			RefOType type = RefOType.valueOf(attributes.getNamedItem("type").getNodeValue());
			ReferentialObjectBuilder builder = new ReferentialObjectBuilder(identifier, type);
			log.trace("creating referentialObject " + identifier + " of type " + type);
			Node defIdentNode = attributes.getNamedItem("name");
			Node genderNode = attributes.getNamedItem("gender");
			if (defIdentNode != null) {
				builder.setDefiniteIdentification(defIdentNode.getNodeValue());
				log.trace("found definite identification " + defIdentNode.getNodeValue() + ", assigning it to " + identifier);
			}
			if (genderNode != null) {
				builder.setGender(Gender.valueOf(genderNode.getNodeValue()));
				log.trace("found gender " + genderNode.getNodeValue() + ", assigning it to " + identifier);
			}
			refOs.put(identifier, builder);
		}
		log.debug("done creating referential objects, starting to assign attributes");
		for (int i = 0; i < nl.getLength(); i++) {
			Node refOnode = nl.item(i);
			if (!refOnode.getNodeName().equals("vavetam:referential-object")) {
				continue;
			}
			NodeList subnodes = refOnode.getChildNodes();
			for (int j = 0; j < subnodes.getLength(); j++) {
				Node subnode = subnodes.item(j);
				if (subnode.getNodeName().equals("vavetam:subparts")) {
					handleSubpartsNode(refOs, subnode);
				}
				if (subnode.getNodeName().equals("vavetam:superparts")) {
					handleSuperpartsNode(refOs, subnode);
				}
				if (subnode.getNodeName().equals("vavetam:descriptor")) {
					handleDescriptorNode(refOs, subnode);
				}
			}
		}
		log.debug("done assigning attributes, start finalizing referential objects");
		Map<String, IReferentialObject> result = new HashMap<String, IReferentialObject>();
		for (Entry<String, ReferentialObjectBuilder> e : refOs.entrySet()) {
			result.put(e.getKey(), e.getValue().result());
		}
		log.debug("done finalizing referential objects and parsing the xml file");
		return Collections.unmodifiableMap(result);
	}

	private void handleSubpartsNode(Map<String, ReferentialObjectBuilder> refOs, Node node) {
		String id = node.getParentNode().getAttributes().getNamedItem("id").getNodeValue();
		ReferentialObjectBuilder builder = refOs.get(id);
		NodeList partnodes = node.getChildNodes();
		for(int i = 0; i < partnodes.getLength(); i++) {
			Node partnode = partnodes.item(i);
			if (partnode.getNodeName().equals("vavetam:part")) {
				String ref = partnode.getAttributes().getNamedItem("ref").getNodeValue();
				builder.addChild(refOs.get(ref).result());
			}
		}
		log.trace("added subparts " + builder.childrenToString() + " to " + builder.getIdentifier());
	}

	private void handleSuperpartsNode(Map<String, ReferentialObjectBuilder> refOs, Node node) {
		String id = node.getParentNode().getAttributes().getNamedItem("id").getNodeValue();
		ReferentialObjectBuilder builder = refOs.get(id);
		NodeList partnodes = node.getChildNodes();
		for(int i = 0; i < partnodes.getLength(); i++) {
			Node partnode = partnodes.item(i);
			if (partnode.getNodeName().equals("vavetam:part")) {
				String ref = partnode.getAttributes().getNamedItem("ref").getNodeValue();
				builder.addParent(refOs.get(ref).result());
			}
		}
		log.trace("added superparts " + builder.parentsToString() + " to " + builder.getIdentifier());
	}
	
	private void handleDescriptorNode(Map<String, ReferentialObjectBuilder> refOs, Node node) {
		Definiteness def= Definiteness.valueOf(node.getAttributes().getNamedItem("definiteness").getNodeValue());
		IDescriptor result = new Descriptor(def);
		String id = node.getParentNode().getAttributes().getNamedItem("id").getNodeValue();
		ReferentialObjectBuilder builder = refOs.get(id);
		NodeList predicateNodes = node.getChildNodes();
		for (int i = 0; i < predicateNodes.getLength(); i++) {
			Node predicateNode = predicateNodes.item(i);
			if (predicateNode.getNodeName().equals("vavetam:predicate")) {
				PredicateName name = PredicateName.valueOf(predicateNode.getAttributes().getNamedItem("name").getNodeValue());
				List<IReferentialObject> args = new ArrayList<IReferentialObject>();
				NodeList argNodes = predicateNode.getChildNodes();
				for (int j = 0; j < argNodes.getLength(); j++) {
					Node argNode = argNodes.item(j);
					if (argNode.getNodeName().equals("vavetam:arg")) {
						String ref = argNode.getAttributes().getNamedItem("ref").getNodeValue();
						if (ref.equals("X")) {
							args.add(builder.result());
						} else {
							args.add(refOs.get(ref).result());
						}
					}
				}
				IPredicate predicate = new Predicate(name, args);
				result.add(predicate);
			}
		}
		builder.addDescriptor(result);
		log.trace("added descriptor " + result + " to " + builder.getIdentifier());
	}
}
