package vavetam.implementation.generation.incremental;

import java.text.MessageFormat;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.apache.log4j.Logger;

import com.google.common.collect.Lists;
import com.google.common.collect.Queues;

/**
 * A pipeline processor is a concurrent and type safe working unit of a bigger pipeline.
 * Pipeline processors can easily be fitted together and every pipeline processor can
 * have multiple others listening to its output.
 * 
 * 
 * @author Ole Eichhorn
 *
 * @param <I> the input type to the pipeline processor
 * @param <O> the output type of the pipeline processor
 */
public abstract class AbstractPipelineProcessor<I, O> implements Runnable {
	
	
	private List<AbstractPipelineProcessor<O, ? extends Object>> listeners = Lists.newArrayList();
	private ConcurrentLinkedQueue<I> input = Queues.newConcurrentLinkedQueue();
	private Thread worker = new Thread(this);

	//flags
	private boolean running = false;
	private boolean finish = false;
	
	//properties
	private int postProcessDelay = 0;
	private int repetitions = 1;
	private int maximumQueueLength = 1000 * 1000;
	private OverflowPolicy overflowPolicy= OverflowPolicy.DROP_SILENTLY;

	protected abstract O process(I in);
	
	protected String threadName() {
		return "Pipeline Processor";
	}
	
	/**
	 * register a pipeline processor to listen to the output of this one.
	 * the input type of that processor has to match the output type of this one.
	 * @param listener
	 */
	public void registerListener(AbstractPipelineProcessor<O, ? extends Object> listener) {
		listeners.add(listener);
	}
	
	/**
	 * 
	 * @param in the new input to the processor
	 * @throws Exception if the {@link OverflowPolicy} is set to {@link OverflowPolicy.THROW_EXCEPTION}
	 * 		   and the input queue exceeds {@code getMaximumQueueLength()} items
	 */
	public void put (I in) throws Exception {
		if (input.size() < maximumQueueLength) {
			input.add(in);
			synchronized (this) { notify(); }
		} else {
			switch (overflowPolicy) {
			case REFUSE: throw new RuntimeException("OVERFLOW! cannot add " + in);
			case REPLACE_OLDEST: { input.poll(); input.add(in); }
			case DROP_SILENTLY: getLogger().trace(MessageFormat.format("dropping input {0} due to queue overflow!", in));
			}
		}
	}
	
	@Override
	public void run() {
		running = true;
		while (running) {
			if (input.isEmpty()) {
				if (finish) {
					getLogger().trace("finishing now");
					stop();
				} else try {
					getLogger().trace("starting to wait for input");
					synchronized(this) { wait(); }
				} catch (Exception e) {
					getLogger().trace("been interrupted, perhaps input arrived. continuing");
				}
				continue;
			}
			getLogger().trace("processing next input");
			O out = process(input.poll());
			for (int i = 0; i < repetitions; i++) {
				if (postProcessDelay > 0) {
					getLogger().trace(MessageFormat.format("waiting {0}ms before propagating output", postProcessDelay));
					try {
						Thread.sleep(postProcessDelay);
					} catch (InterruptedException e) {
						//oops!
					}
				}
				getLogger().trace(MessageFormat.format((repetitions > 1) ? "propagating output, round {0}" : "propagating output", i+1));
				for (AbstractPipelineProcessor<O, ? extends Object> listener : listeners) {
					try {
						listener.put(out);
					} catch (Exception e) {
						handleOverflow(out, listener);
					}
				}
			}
		}
	}

	/**
	 * @param millis the delay after which a processing result is to be passed 
	 * on to the listeners
	 */
	public void setPostProcessDelay(int millis) {
		postProcessDelay = millis;
	}
	
	/**
	 * @param repetitions the amount of times, an output is to be sent to the
	 * listeners
	 */
	public void setRepetitions(int repetitions) {
		this.repetitions = repetitions;
	}
	
	/**
	 * Makes this pipeline processor start to work.
	 */
	public void start() {
		worker.setName(threadName());
		worker.start();
	}
	
	/**
	 * Makes this pipeline processor stop working at the next possible opportunity.
	 */
	public void stop() {
		running = false;
	}
	
	/**
	 * Tells this pipeline processor to stop working as soon as the end of its
	 * input queue is reached. Also, all listeners are requested to do this
	 * recursively.
	 */
	public void propagateFinishAndStop() {
		finishAndStop();
		for (AbstractPipelineProcessor<O, ? extends Object> listener: listeners) {
			listener.propagateFinishAndStop();
		}
	}
	
	/**
	 * Makes this pipeline processor stop working at the next possible opportunity.
	 * Also, all listeners are requested to do this recursively.
	 */
	public void propagateStop() {
		stop();
		for (AbstractPipelineProcessor<O, ? extends Object> listener: listeners) {
			listener.propagateStop();
		}
	}
	
	/**
	 * Tells this pipeline processor to stop working as soon as the end of its
	 * input queue is reached
	 */
	public void finishAndStop() {
		this.finish = true;
		synchronized(this) {
			notify();
		}
	}
	
	/**
	 * Make the working thread of this pipeline processor a daemon thread.
	 */
	public void setDaemon() {
		worker.setDaemon(true);
	}
	
	/**
	 * Set the maximum number of items in the input queue of this pipeline
	 * processor. if this number is exceeded, an overflow event is caused.
	 * @param length the maximum length of the queue.
	 * @see OverflowPolicy
	 * @see getMaximumQueueLength
	 */
	public void setMaximumWaitingQueueLength(int length) {
		maximumQueueLength = length;
	}
	
	/**
	 * @return the maximum number of items in the input queue of this pipeline
	 * processor.
	 * @see OverflowPolicy
	 * @see getMaximumQueueLength
	 */
	public int getMaximumQueueLength() {
		return maximumQueueLength;
	}
	
	/**
	 * Defines how a pipeline processor handles input if its queue is full
	 * 
	 * @author Ole Eichhorn
	 *
	 */
	public enum OverflowPolicy {
		/**
		 * drops input without any notification if the queue is full
		 */
		DROP_SILENTLY("drop the item silently"),
		
		/**
		 * refuses and hands back input to the issuer if the queue is full
		 */
		REFUSE("refuse and hand back the input"),
		
		/**
		 * removes the oldest input item in favor for the new one
		 */
		REPLACE_OLDEST("remove the oldest input item and put the new one in");
		
		private String description;
		
		private OverflowPolicy(String description) {
			this.description = description;
		}
		
		/**
		 * 
		 * @return a human readable description for the {@link OverflowPolicy}
		 */
		public String getDescription() {
			return description;
		}
	}
	
	/**
	 * This method is called if a listener refuses to accept input because its
	 * input queue is full. 
	 * 
	 * @param out the output that has been refused
	 * @param refuser the listener that has refused the input
	 */
	protected void handleOverflow(O out, AbstractPipelineProcessor<O, ? extends Object> refuser) {
		// standard implementation does nothing;
	}
	
	public void setOverflowPolicy(OverflowPolicy policy) {
		overflowPolicy = policy;
	}
	
	protected Logger getLogger() {
		return Logger.getLogger(AbstractPipelineProcessor.class);
	}
	
	protected boolean inputIsEmpty() {
		return input.isEmpty();
	}
	
	protected void clearInput() {
		input.clear();
	}
}