package vavetam.implementation.generation.incremental;

import static java.text.MessageFormat.format;
import inpro.apps.SimpleMonitor;
import inpro.incremental.processor.SynthesisModule;
import inpro.incremental.unit.EditMessage;
import inpro.incremental.unit.EditType;
import inpro.incremental.unit.IU;
import inpro.incremental.unit.IU.IUUpdateListener;
import inpro.incremental.unit.PhraseIU;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.Predicate;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import vavetam.api.structure.IFormulation;
import vavetam.api.structure.IMapExploratoryProcedure;
import vavetam.api.structure.IPreverbalMessage;
import vavetam.implementation.generation.incremental.formulation.ExtensionFormulator;
import vavetam.implementation.generation.incremental.formulation.GeometricRelationsFormulator;
import vavetam.implementation.generation.incremental.formulation.IFormulator;
import vavetam.implementation.generation.incremental.formulation.IdentificationFormulator;
import vavetam.implementation.generation.incremental.formulation.JunctionsFormulator;
import vavetam.implementation.generation.incremental.formulation.SpatialRelationsFormulator;
import vavetam.implementation.generation.incremental.heuristic.DifferenceFirstPreference;
import vavetam.implementation.simulation.DummySpeechEngine;
import vavetam.implementation.structure.Formulation;
import vavetam.implementation.structure.enums.Intention;


import com.google.common.collect.Sets;

public class FormulationModule extends AbstractPipelineProcessor<IMapExploratoryProcedure, Object> {
	
	private Logger log = Logger.getLogger(FormulationModule.class);
	private Map<Set<IPreverbalMessage>, List<IFormulation>> formulationsForPVMs = new HashMap<Set<IPreverbalMessage>, List<IFormulation>>();
	private Map<PhraseIU, IFormulation> formulationsForPhraseIUs = new HashMap<PhraseIU, IFormulation>();
	private Map<Intention, IFormulator> formulators = new HashMap<Intention, IFormulator>();
	private boolean dummyMode = false;
	private boolean incrementalMode = true;
	
	boolean nonIncrementalSynthesisRunning = false;

	/**
	 * diese Queue sollte wiederspiegeln, was die Synthese gerade und in Zukunft sagt (nach aktuellem Stand)
	 */
	private ConcurrentLinkedQueue<IPreverbalMessage> lastFormulatedPVMs = new ConcurrentLinkedQueue<IPreverbalMessage>();
	/**
	 * diese Queue sollte wiederspiegeln, was die Synthese gerade und in Zukunft sagt (nach aktuellem Stand)
	 */
	private ConcurrentLinkedQueue<List<PhraseIU>> lastFormulatedUtterances = new ConcurrentLinkedQueue<List<PhraseIU>>();
	
	private SynthesisModule synthesis;
	
	public FormulationModule(AbstractPipelineProcessor<? extends Object, IMapExploratoryProcedure> predecessor, String dummyMode, String incrementalMode) {
		predecessor.registerListener(this);
		setDummyMode(dummyMode);
		setIncrementalMode(incrementalMode);
		//create formulators
		formulators.put(Intention.IDENTIFICATION, new IdentificationFormulator());
		formulators.put(Intention.SPATIAL_RELATIONS, new SpatialRelationsFormulator()); 
		formulators.put(Intention.EXTENSION, new ExtensionFormulator());
		formulators.put(Intention.JUNCTIONS, new JunctionsFormulator());
		formulators.put(Intention.GEOMETRIC_RELATIONS, new GeometricRelationsFormulator());
		start();
	}

	@Override
	public Object process(IMapExploratoryProcedure mep) {
		if (incrementalMode) {
			processIncremental(mep);
		} else {
			processNonIncremental(mep);
		}
		return null;
	}
	

	private void processIncremental(IMapExploratoryProcedure mep) {
		if (mep.getPreverbalMessages().isEmpty()) {
			log.trace(format("no PVMs in {0}, cannot formulate anything!", mep));
			return;
		}
		if (isSomethingOngoing()) {
			// eine MEP wird gerade noch verarbeitet
			log.debug(format("{0} is still ongoing", lastFormulatedPVMs.peek()));

			IFormulation currentFormulation = formulationsForPhraseIUs.get(getCurrentlyUtteredPhraseIU());

			Set<IPreverbalMessage> currentPVMs = Collections.emptySet();
			if (currentFormulation != null) {
				log.trace(format("current formulation is {0}", currentFormulation));
				currentPVMs = currentFormulation.getPvms();
			} else {
				log.trace(format("no current formulation found! current PhraseIU: {0}, available mappings: {1}", getCurrentlyUtteredPhraseIU(), StringUtils.join(formulationsForPhraseIUs.entrySet(), ", ")));
			}
			if (currentPVMs.isEmpty()) {
				//do nothing
				log.debug("no PVMs are currently being uttered so no adaption is possible");
			} else  if (currentPVMs.size() > 1){
				//es werden schon mehr als eine PVM wiedergegeben
				//TODO: was machen wir hier??
				log.debug("cant adapt since there's more than one pvm being uttered already");
			} else {
				//die einzige aktuell wiedergegebene PVM wird versucht, mit der ersten neuen zu kombinieren
				log.debug(format("trying to adapt ongoing utterance {0}", currentFormulation));
				IPreverbalMessage currentPVM = currentPVMs.iterator().next();
				IPreverbalMessage newPVM = mep.getPreverbalMessages().get(0);
				List<IFormulation> combinedFormulations = formulate(currentPVM, newPVM);
				final IFormulation current = currentFormulation;
				log.trace(format("there are {0} messages before filtering", combinedFormulations.size()));
				// filtere Formulierungen, die sich eignen, um zu adaptieren.
				CollectionUtils.filter(combinedFormulations, new Predicate(){

					@Override
					public boolean evaluate(Object arg0) {
						if (arg0 instanceof IFormulation) {
							return canAdapt(current, ((IFormulation) arg0));
						}
						return false;
					}

				});

				if (combinedFormulations.size() > 0) {
					// zugunsten der Variation
					Collections.shuffle(combinedFormulations, new Random(System.currentTimeMillis()));
					// sortieren nach "geringste ähnlichkeit zuerst"
					Collections.sort(combinedFormulations, new DifferenceFirstPreference(current));
					log.trace(format("formulated {0} combined utterance possibilities for {1} and {2}", combinedFormulations.size(), currentPVM, newPVM));
					IFormulation formulation = combinedFormulations.get(0);
					log.debug(format("adapting to {0}.", formulation));
					doAdapt(currentFormulation, formulation);
					formulationsForPVMs.put(Sets.newHashSet(currentPVM, newPVM), combinedFormulations);
					// die erste PVM wurde damit abgearbeitet und wird entfernt;
					IPreverbalMessage pvm = mep.getPreverbalMessages().remove(0);
					lastFormulatedPVMs.add(pvm);
				} else {
					log.debug("it's been too late, no adaption possible. continuing with regular formulation process.");
				}
			}
		}
		
		for (IPreverbalMessage pvm: mep.getPreverbalMessages()) {
			List<IFormulation> messages = formulate(pvm);
			if (messages.isEmpty()){
				log.trace("no possibilities, so doing nothing");
				return;
			}
			log.trace(format("formulated utterance possibilities {0} for {1}", messages, pvm));
			formulationsForPVMs.put(Collections.singleton(pvm), messages);
			Collections.shuffle(messages, new Random(System.currentTimeMillis()));
			IFormulation formulation = messages.get(0); // best formulation available
			log.trace("selected best possible utterance " + formulation);
			final List<PhraseIU> ius = formulation.getIus();
			for (PhraseIU piu: ius) {
				formulationsForPhraseIUs.put(piu, formulation);
			}
			
			installUpdateHooks(ius);
			
			List<PhraseIU> adds = new ArrayList<PhraseIU>();
			List<EditMessage<PhraseIU>> edits= new ArrayList<EditMessage<PhraseIU>>();
			//adde alle neuen ius
			for (PhraseIU iu: ius) {
				adds.add(iu);
				edits.add(new EditMessage<PhraseIU>(EditType.ADD, iu));
			}
			log.trace("updating reference queues");
			lastFormulatedPVMs.add(pvm);
			log.trace(format("new status of pvm reference queue: {0}", lastFormulatedPVMs));
			lastFormulatedUtterances.add(adds);
			log.trace(format("new status of utterances reference queue: {0}", lastFormulatedUtterances));
			synthesis.hypChange(ius, edits);
		}
	}
	
	private void processNonIncremental(IMapExploratoryProcedure mep) {
		for (IPreverbalMessage pvm: mep.getPreverbalMessages()) {
			List<IFormulation> messages = formulate(pvm);
			if (messages.isEmpty()){
				log.trace("no possibilities, so doing nothing");
				return;
			}
			log.trace(format("formulated utterance possibilities {0} for {1}", messages, pvm));
			formulationsForPVMs.put(Collections.singleton(pvm), messages);
			IFormulation formulation = messages.get(0); // best formulation available
			log.trace("selected best possible utterance " + formulation);
			final List<PhraseIU> ius = formulation.getIus();
			for (PhraseIU piu: ius) {
				formulationsForPhraseIUs.put(piu, formulation);
			}
			
			installUpdateHooks(ius);
			
			List<PhraseIU> adds = new ArrayList<PhraseIU>();
			List<EditMessage<PhraseIU>> edits= new ArrayList<EditMessage<PhraseIU>>();
			//adde alle neuen ius
			for (PhraseIU iu: ius) {
				adds.add(iu);
				edits.add(new EditMessage<PhraseIU>(EditType.ADD, iu));
			}
			log.trace("updating reference queues");
			lastFormulatedPVMs.add(pvm);
			log.trace(format("new status of pvm reference queue: {0}", lastFormulatedPVMs));
			lastFormulatedUtterances.add(adds);
			log.trace(format("new status of utterances reference queue: {0}", lastFormulatedUtterances));
			// künstlich blocken, bis synthese durch ist...
			adds.get(adds.size()-1).addUpdateListener(new IUUpdateListener() {
				
				@Override
				public void update(IU updatedIU) {
					if (updatedIU.isCompleted()) {
						nonIncrementalSynthesisRunning = false;
						synchronized(this) {
							this.notify();
						}
					}
				}
			});
			nonIncrementalSynthesisRunning = true;
			synthesis.hypChange(adds, edits);
			log.trace("waiting for non-incremental synthesis to finish");
			synchronized(this) {
				while (nonIncrementalSynthesisRunning) 
					try {
						wait(50);
					} catch (InterruptedException e) {
						// block on
					}
			}
			log.trace("non-incremental synthesis has finished, moving on");
			if (!inputIsEmpty()) {
				break;
			}
		}
	}
	
	private List<IFormulation> formulate(IPreverbalMessage... pvms) {
		if (pvms.length == 0) {
			log.debug("no pvms to formulate specified!");
			return Collections.emptyList();
		} else {
			Intention intention = pvms[0].getIntention();
			for (int i = 0; i < pvms.length; i++) {
				IFormulator formulator = formulators.get(intention);
				if (formulator == null) {
					log.debug(format("no Formulator for {0} defined", intention));
					continue;
				}
				List<IFormulation> formulations = formulator.formulate(pvms);
				if (!formulations.isEmpty()) {
					return formulations;
				}
			}
		}
		log.debug("tried all available formulators, achieved no result. returning empty list.");
		return Collections.emptyList();
	}
	
	protected String threadName() {
		return "Formulation Module";
	}
	
	@Override
	public void stop() {
		super.stop();
		synthesis.finish();
	}
	
	private PhraseIU getCurrentlyUtteredPhraseIU() {
		log.trace("trying to find currently Uttered PhraseIU. currentUtterances: [" + StringUtils.join(lastFormulatedUtterances.peek(), ", ") + "]");			
		for (PhraseIU iu: lastFormulatedUtterances.peek()) {
			if (iu.isOngoing()) {
				return iu;
			}
		}
		return null;
	}
	
	private boolean canAdapt(IFormulation oldFormulation, IFormulation newFormulation) {
		int progressIndex = findProgress(oldFormulation);
		if (progressIndex == -1) {
			log.trace("no progress in old formulation, cancelling.");
			return false;
		}
		PhraseIU progress = oldFormulation.getIus().get(progressIndex);
		if (progress == null) {
			log.warn("couldn't find current PhraseIU, this shouldn't happen!");
			return false;
		}
		Iterator<PhraseIU> oldIUs = oldFormulation.getIus().iterator();
		Iterator<PhraseIU> newIUs = newFormulation.getIus().iterator();
		
		while(oldIUs.hasNext()) {
			if (newIUs.hasNext()) {
				PhraseIU oldIU = oldIUs.next();
				PhraseIU newIU = newIUs.next();
				if (!oldIU.toPayLoad().equals(newIU.toPayLoad())) {
					return false;
				}
				if (oldIU == progress) {
					break;
				}
			} else {
				return false;
			}
		}
		log.trace(format("{0} has progressed to {1}, so it can adapt to {2}", oldFormulation, progress, newFormulation));
		return true;
	}

	private int findProgress(IFormulation formulation) {
		int result = 0;
		for(PhraseIU iu: formulation.getIus()) {
			if (iu.isOngoing()) {
				return result;
			}
			result ++ ;
		}
		return -1;
	}
	
	private void doAdapt(IFormulation oldFormulation, IFormulation newFormulation) {
		// ermittle fortschritt:
		int progressIndex = findProgress(oldFormulation);
		List<EditMessage<PhraseIU>> edits = new ArrayList<EditMessage<PhraseIU>>();
		// entferne veraltetes
		clearUnsaidIUs();
		//adde alle neuen ius
		List<PhraseIU> newIUs = newFormulation.getIus();
		List<PhraseIU> adds = new ArrayList<PhraseIU>();
		for (int i = progressIndex + 1; i < newIUs.size(); i++) {
			PhraseIU iu = newIUs.get(i);
			log.trace("adding " + iu + "to sythesis buffer.");
			adds.add(iu);
			edits.add(new EditMessage<PhraseIU>(EditType.ADD, newIUs.get(i)));
		}
		installUpdateHooks(adds);
		// utterance reference queue neu bestücken
		lastFormulatedUtterances.add(adds);
		
		synthesis.hypChange(adds, edits);
	}
	
	public void setDummyMode(String mode) {
		dummyMode = Boolean.valueOf(mode);
		if (dummyMode) {
			log.debug("using dummy mode for the speech engine! change in vavetam.properties");
			synthesis = new DummySpeechEngine();
		} else {
			synthesis = new SynthesisModule(SimpleMonitor.setupDispatcher());
		}
		Formulation.setDummyMode(dummyMode);
	}
	
	private boolean isSomethingOngoing() {

		List<IFormulation> currentFormulations = formulationsForPVMs.get(Collections.singleton(lastFormulatedPVMs.peek()));
		
		boolean completed = (currentFormulations == null) ? true : currentFormulations.get(0).isCompleted();
		
		return !lastFormulatedPVMs.isEmpty() && !completed;
	}
	
	private void installUpdateHooks(final List<PhraseIU> ius ){
		if (ius.size() > 1) {
			ius.get(ius.size()-2).addUpdateListener(new IUUpdateListener() {
				//kurz vorher committen
				@Override
				public void update(IU updatedIU) {
					if (updatedIU.isCompleted()) {
						PhraseIU phrase = ius.get(ius.size()-1);
						List<IU> ius = Collections.emptyList();
						List<EditMessage<PhraseIU>> edit = Collections.singletonList(new EditMessage<PhraseIU>(EditType.COMMIT, phrase));
						
						synthesis.hypChange(ius, edit);
					}
				}
			});
		}
		
		if (ius.size() > 0) {
			ius.get(ius.size()-1).addUpdateListener(new IUUpdateListener() {
				// aktuelle PVM für Determinierungsprozess hier im FormulationModule updaten 
				@Override
				public void update(IU updatedIU) {
					if (updatedIU.isCompleted()) {
						//auf diese Weise ist der head der Queues immer auch das aktuell wiedergegebene.
						log.trace("updating reference queues");
						lastFormulatedPVMs.poll();
						log.trace(format("new status of pvm reference queue: {0}", lastFormulatedPVMs));
						lastFormulatedUtterances.poll();
						log.trace(format("new status of utterances reference queue: {0}", lastFormulatedUtterances));
					}
				}
			});
		}
		
		if (!incrementalMode) {
			ius.get(ius.size()-1).addUpdateListener(new IUUpdateListener() {
				
				@Override
				public void update(IU updatedIU) {
					if (updatedIU.isCompleted()) {
						nonIncrementalSynthesisRunning = false;
						synchronized(this) {
							this.notify();
						}
					}
				}
			});
		}
	}

	@Override
	protected Logger getLogger() {
		return log;
	}
	
	@SuppressWarnings("unchecked")
	private void clearUnsaidIUs() {
		List<EditMessage<PhraseIU>> edits = new ArrayList<EditMessage<PhraseIU>>();
		// revoke noch nicht gesagtes
		for (List<PhraseIU>  ius : lastFormulatedUtterances) {
			for (PhraseIU iu : ius) {
				if (iu.isUpcoming()) {
					log.trace(format("revoking {0} from synthesis buffer.",iu));
					edits.add(new EditMessage<PhraseIU>(EditType.REVOKE, iu));
				}
			}
		}
		synthesis.hypChange(Collections.EMPTY_SET, edits);
		// reference queues leeren
		lastFormulatedPVMs.clear();
		lastFormulatedUtterances.clear();
	}

	public void setIncrementalMode(String mode) {
		incrementalMode = Boolean.parseBoolean(mode);
		if (incrementalMode) {
			setMaximumWaitingQueueLength(1);
			setOverflowPolicy(OverflowPolicy.REPLACE_OLDEST);
		} else {
			setMaximumWaitingQueueLength(1000 * 1000);
			setOverflowPolicy(OverflowPolicy.DROP_SILENTLY);
		}
		log.info("incremental mode is switched " + (incrementalMode ? "on" : "off"));
	}
}