package vavetam.implementation.generation.incremental;

import static java.text.MessageFormat.format;


import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Stack;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;

import vavetam.api.structure.IMapExploratoryProcedure;
import vavetam.api.structure.IReferentialObject;
import vavetam.implementation.structure.UtterancePlan;


import com.google.common.collect.Lists;

public class History {

	private Stack<Pair<IMapExploratoryProcedure, UtterancePlan>> memory = new Stack<Pair<IMapExploratoryProcedure,UtterancePlan>>();
	private Logger log = Logger.getLogger(History.class);
	
	public Difference calculateDifferenceToLast(IMapExploratoryProcedure mep) {
		IMapExploratoryProcedure last = null;
		if (!memory.isEmpty()) {
			last = memory.peek().getLeft();
			log.trace(format("comparing {0} to {1}", mep, last));
		} else {
			log.trace(format("first memory item: {0} not comparing to anything", mep));
		}
		DifferenceType type;
		List<IReferentialObject> affectedObjects;
		if (last == null || !last.primaryFocusEquals(mep)) {
			log.trace("so this is a primary focus change...");
			log.debug("primary focus change detected");
			type = DifferenceType.PRIMARY_FOCUS_CHANGE;
			affectedObjects = Lists.newArrayList(mep.getPrimaryFocus());
			if (last != null) {
				affectedObjects.add(last.getPrimaryFocus());
			}
			
		} else {
			log.trace("nothing in the primary focus changed, so lets have a look at the secondray...");
			List<IReferentialObject> inThisButNotInLast = new ArrayList<IReferentialObject>(mep.getSecondaryFocus());
			inThisButNotInLast.removeAll(last.getSecondaryFocus());
			log.trace("in this, but not in last secondary focus: " + StringUtils.join(inThisButNotInLast, ", "));
			List<IReferentialObject> inLastButNotInThis = new ArrayList<IReferentialObject>(last.getSecondaryFocus());
			inLastButNotInThis.removeAll(mep.getSecondaryFocus());
			log.trace("in last, but not in this secondary focus: " + StringUtils.join(inLastButNotInThis, ", "));
			if (!inThisButNotInLast.isEmpty() && inLastButNotInThis.isEmpty()) {
				log.debug("secondary focus change detected");
				log.trace("new secondary foci added: " + StringUtils.join(inThisButNotInLast, ", "));
				type = DifferenceType.SECONDARY_FOCUS_ADDED;
				affectedObjects = inThisButNotInLast;
			} else if (!inLastButNotInThis.isEmpty() && inThisButNotInLast.isEmpty()) {
				log.debug("secondary focus change detected");
				log.trace("old secondary foci removed: " + StringUtils.join(inLastButNotInThis, ", "));
				type = DifferenceType.SECONDARY_FOCUS_REMOVED;
				affectedObjects = inLastButNotInThis;
			} else if (!inThisButNotInLast.isEmpty() && !inLastButNotInThis.isEmpty()) {
				log.debug("secondary focus change detected");
				log.trace(format("new secondary foci added: {0} and old removed: {1}", StringUtils.join(inThisButNotInLast, ", "), StringUtils.join(inLastButNotInThis, ", ")));
				type = DifferenceType.SECONDARY_FOCUS_ADDED_AND_REMOVED;
				affectedObjects = inThisButNotInLast;
				affectedObjects.addAll(inLastButNotInThis);
			} else {
				log.debug("no change detected");
				type = DifferenceType.NO_CHANGE;
				affectedObjects = Collections.emptyList();
			}
		}
		return new Difference(type, affectedObjects);
	}
	
	public void memorize(IMapExploratoryProcedure mep, UtterancePlan iu) {
		log.trace(format("memorizing {0} with {1} at {2}", mep, iu, "[implement time stamp]"));
		memory.push(new ImmutablePair<IMapExploratoryProcedure, UtterancePlan>(mep, iu));
	}
	
	public boolean hadThat(IMapExploratoryProcedure mep, UtterancePlan iu) {
		log.trace(format("looking up history for {0}", mep));
		boolean result = (memory.contains(new ImmutablePair<IMapExploratoryProcedure, UtterancePlan>(mep, iu)));
		log.trace(format("{0} has{1} been mentioned before", mep, result ? "" : " not"));
		return result;
	}
	
	public class Difference {
		private final DifferenceType type;
		private final List<IReferentialObject> affectedObjects;
		
		public Difference(DifferenceType type, List<IReferentialObject> affectedObjects) {
			this.type = type;
			this.affectedObjects = affectedObjects;
		}
		
		public DifferenceType getType() {
			return type;
		}
		
		public List<IReferentialObject> getAffectedObjects() {
			return affectedObjects;
		}
	}

	public enum DifferenceType {
		PRIMARY_FOCUS_CHANGE,
		SECONDARY_FOCUS_ADDED,
		SECONDARY_FOCUS_REMOVED,
		SECONDARY_FOCUS_ADDED_AND_REMOVED,
		NO_CHANGE
	}
}
