package vavetam.implementation.generation.incremental.formulation;

import static java.text.MessageFormat.format;
import static vavetam.implementation.generation.incremental.formulation.Dictionary.ArticleDefiniteness.DEFINITE;
import static vavetam.implementation.generation.incremental.formulation.Dictionary.ArticleDefiniteness.INDEFINITE;
import static vavetam.implementation.generation.incremental.formulation.Dictionary.Casus.DATIV;
import static vavetam.implementation.generation.incremental.formulation.Dictionary.Casus.NOMINATIV;
import static vavetam.implementation.generation.incremental.formulation.Dictionary.Gender.FEMALE;
import static vavetam.implementation.generation.incremental.formulation.Dictionary.Gender.MALE;
import static vavetam.implementation.generation.incremental.formulation.Dictionary.Gender.NEUTRAL;
import static vavetam.implementation.structure.util.PredicateUtils.isIndefiniteIdentificationPredicate;

import java.util.LinkedList;
import java.util.List;
import java.util.ResourceBundle;

import org.apache.commons.lang3.StringUtils;

import vavetam.api.structure.IDescriptor;
import vavetam.api.structure.IPreverbalMessage;
import vavetam.api.structure.IReferentialObject;
import vavetam.implementation.structure.enums.PredicateName;



/**
 * this class is used by the Articulator to construct valid German sentences
 *
 */
public class Dictionary {
	
	private static final String BUNDLE_NAME = "vavetam.implementation.generation.incremental.formulation.dictionary"; //$NON-NLS-1$
	private static final ResourceBundle BUNDLE = ResourceBundle.getBundle(BUNDLE_NAME);
	
	public static final String WHITESPACE = " "; //$NON-NLS-1$
	public static final String PERIOD = "."; //$NON-NLS-1$
	public static final String COMMA = ","; //$NON-NLS-1$
	public static final String AGAIN = BUNDLE.getString("AGAIN"); //$NON-NLS-1$
	public static final String AND = BUNDLE.getString("AND"); //$NON-NLS-1$
	public static final String IS = BUNDLE.getString("IS"); //$NON-NLS-1$
	public static final String ARE = BUNDLE.getString("ARE"); //$NON-NLS-1$
	public static final String HESITATION = BUNDLE.getString("HESITATION"); //$NON-NLS-1$
	
	private static final String ARTICLE_KEY = "Article_{0}_{1}_{2}"; //$NON-NLS-1$
	private static final String PERSPRONOM_KEY = "PersPronom_{0}_{1}"; //$NON-NLS-1$
	private static final String ORIENTATION1_KEY = "Orientation1_{0}"; //$NON-NLS-1$
	private static final String ORIENTATION2_KEY = "Orientation2_{0}"; //$NON-NLS-1$
	private static final String DESCR_KEY = "Descr_{0}"; //$NON-NLS-1$
	private static final String IDENT_KEY = "Ident_{0}"; //$NON-NLS-1$
	private static final String IDENT_COMPOUND = "Ident_Compound_"; //$NON-NLS-1$
	private static final String IDENT_COMPOUND_BEGIN = IDENT_COMPOUND + "Begin_{0}"; //$NON-NLS-1$
	private static final String IDENT_COMPOUND_END = IDENT_COMPOUND + "End_{0}"; //$NON-NLS-1$
	private static final String GENDER_KEY = "Gender_{0}"; //$NON-NLS-1$
	
	/**
	 * for the purpose of matching a String against a number of other Strings. <br>
	 * Apache's {@link StringUtils#containsAny(CharSequence, CharSequence)} can't check against multiple Strings. Regex magic to the rescue! <br>
	 * see http://stackoverflow.com/a/8992179 <br><br>
	 * 
	 *  this translates as follows:<br>
	 *  -beginning of line(^), <br>
	 *  -zero or more occurrences of any character with reluctant quantifier (.*?), <br>
	 *  -exactly one of the possibilities substituted for {0} (those have to be separated by | and are substituted by <code>{@link java.text.MessageFormat#format(Object)}</code> below), <br> 
	 *  -zero or more occurrences of any character with greedy quantifier (.*) and <br>
	 *  -end of line($) <br>
	 */ 
	private static final String HINT_REGEX = "^.*?({0}).*$"; //$NON-NLS-1$
	
	private static final String MALE_HINT = format(HINT_REGEX, BUNDLE.getString("MALE_HINTS")); //$NON-NLS-1$
	private static final String FEMALE_HINT = format(HINT_REGEX, BUNDLE.getString("FEMALE_HINTS")); //$NON-NLS-1$
	private static final String NEUTRAL_HINT = format(HINT_REGEX, BUNDLE.getString("NEUTRAL_HINTS")); //$NON-NLS-1$
	
//	private static final Logger log = Logger.getLogger(Dictionary.class);
	
	private Dictionary() {}
	
	public static  String getArticle(ArticleDefiniteness ad, Casus c, Gender g) {
		return BUNDLE.getString(format(ARTICLE_KEY, ad, c, g));
	}
	
	public static String getPersPronom(Casus c, Gender g) {
		return BUNDLE.getString(format(PERSPRONOM_KEY, c, g));
	}
	
	public static String getOrientation1(PredicateName pn) {
		return BUNDLE.getString(format(ORIENTATION1_KEY, pn));
	}
	
	public static String getOrientation2(PredicateName pn) {
		return BUNDLE.getString(format(ORIENTATION2_KEY, pn));
	}
	
	public static String getDescr(PredicateName pn) {
		return BUNDLE.getString(format(DESCR_KEY, pn));
	}
	
	private static Gender getGender(PredicateName pn) {
		return Gender.valueOf(BUNDLE.getString(format(GENDER_KEY, pn)));
	}
	
	private static Gender getGenderByDefIdent(String s) {
		if (s.matches(MALE_HINT)) return MALE;
		if (s.matches(FEMALE_HINT)) return FEMALE;
		if (s.matches(NEUTRAL_HINT)) return NEUTRAL;
		else return null;
	}
	
	public static String formConcatenationSentence(List<String> strings) {
		if (strings.isEmpty()) {
			return StringUtils.EMPTY;
		} else if (strings.size() == 1){
			return strings.get(0);
		} else {
			return new StringBuilder()	.append(StringUtils.join(strings.subList(0, strings.size()-1), COMMA+WHITESPACE))
										.append(WHITESPACE+AND+WHITESPACE)
										.append(strings.get(strings.size()-1))
										.toString();
		}
	}
	
	public static Gender determineGender(IReferentialObject refo) {
		Gender gender = refo.getGender();
		if (gender != null) return gender;
		for (IDescriptor d: refo.getDescriptors()) {
			if (isIndefiniteIdentificationPredicate(d.getPredicateAtIndex(0))) {
				gender = getGender(d.getPredicateAtIndex(0).getName());
				if (gender != null) return gender;
			}
		}
		if (refo.getDefiniteIdentification() != null)
			gender = getGenderByDefIdent(refo.getDefiniteIdentification());
		if (gender != null) return gender;
		return NEUTRAL;
	}
	
	public static String identify(IReferentialObject refo, Casus c) {
		StringBuilder result = new StringBuilder();
		if (refo.getDefiniteIdentification() != null) {
			result.append(getArticle(DEFINITE, c, refo));
			result.append(WHITESPACE);
			result.append(refo.getDefiniteIdentification());
			return result.toString();
		} else {
			result.append(getArticle(INDEFINITE, c, refo));
			for (IDescriptor d: refo.getDescriptors()) {
				if (isIndefiniteIdentificationPredicate(d.getPredicateAtIndex(0))) {						
					result.append(getDescr(d.getPredicateAtIndex(0).getName()));
					break;
				}
			}
		}
		return result.toString();
	}
	
	public static String getArticle(ArticleDefiniteness definite, Casus c, IReferentialObject refo) {
		return getArticle(definite, c, determineGender(refo));
	}
	
	public static String getPersPronom(Casus c, IReferentialObject refo) {
		return getPersPronom(c, determineGender(refo));
	}
	
	public static List<String> createIdentificationMessage(IPreverbalMessage pvm) {
		return createIdentificationMessage(pvm.getReferentialObject(), pvm.isRepetition());
	}
	
	public static List<String> createIdentificationMessage(IReferentialObject refo, boolean repetition) {
		List<String> result = new LinkedList<String>();
		String id;
		String again = repetition ? WHITESPACE+AGAIN : StringUtils.EMPTY;
		for (int i = 0; i < 4; i++) {
			if (i == 3) {
				id = identify(refo, DATIV);
			} else {
				id = identify(refo, NOMINATIV);
			}			
			result.add(format(BUNDLE.getString(format(IDENT_KEY, i)), again, id) + PERIOD);
		}
		return result;
	}
	
	public static List<String> createCompoundIdentificationMessage(IPreverbalMessage pvm1, IPreverbalMessage pvm2) {
		return createCompoundIdentificationMessage(pvm1.getReferentialObject(), pvm2.getReferentialObject(), pvm1.isRepetition(), pvm2.isRepetition());
	}
	
	public static List<String> createCompoundIdentificationMessage(IReferentialObject refo1, IReferentialObject refo2, boolean repetition1, boolean repetition2) {
		List<String> result = new LinkedList<String>();
		for (int begin = 0; begin <= 7; begin++ ) {
			for (int end = 0; end <= 3; end++ ) { 
				StringBuilder formatStringBuilder = new StringBuilder();
				formatStringBuilder	.append(BUNDLE.getString(format(IDENT_COMPOUND_BEGIN, begin)))
									.append(WHITESPACE)
									.append(BUNDLE.getString(format(IDENT_COMPOUND_END, end)))
									.append(PERIOD);
				String formatString = formatStringBuilder.toString();
				String again1 = repetition1 ? WHITESPACE+AGAIN : StringUtils.EMPTY;
				String again2 = repetition2 ? WHITESPACE+AGAIN : StringUtils.EMPTY;
				String id1;
				String id2;
				if (begin == 3 || begin == 7) {
					id1 = identify(refo1, DATIV);
				} else {
					id1 = identify(refo1, NOMINATIV);
				}
				
				if (end == 3) {
					id2 = identify(refo2, DATIV);
				} else {
					id2 = identify(refo2, NOMINATIV);
				}
				
				String message = format(formatString, again1, id1, again2, id2);
				result.add(message);
			}
		}
		for (int begin = 0; begin < 3; begin++) {
			String formatString = new StringBuilder()
				.append(BUNDLE.getString(format(IDENT_KEY, begin)))
				.append(WHITESPACE + HESITATION + WHITESPACE)
				.append("{2}.")
				.toString();
			String id1;
			String id2;
			String again1 = repetition1 ? WHITESPACE+AGAIN : StringUtils.EMPTY;
			if (begin == 3) {
				id1 = getArticle(DEFINITE, DATIV, refo1);
				id2 = identify(refo2, DATIV);
			} else {
				id1 = getArticle(DEFINITE, NOMINATIV, refo1);
				id2 = identify(refo2, NOMINATIV);
			}
			
			result.add(format(formatString, again1, id1, id2));
		}
		return result; 
	}
	
	public static enum Casus {
		NOMINATIV,
		GENITIV,
		AKKUSATIV,
		DATIV,
	}
	
	public static enum Gender {
		MALE,
		FEMALE,
		NEUTRAL
	}
	
	public static enum ArticleDefiniteness {
		DEFINITE,
		INDEFINITE
	}
}
