package vavetam.implementation.simulation;



import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;


import org.apache.log4j.Logger;

import vavetam.implementation.fileio.CSVParser;
import vavetam.implementation.startup.Vavetam;
import vavetam.oldcode.MainFrame;



/**
 * An instance of this class can be used to simulate scheduled input
 * sequences. The schedule used is defined in a CVS file. The Simulator
 * can be controlled through the GUI.
 * 
 * @author Ole Eichhorn
 */
public class AutomaticSimulator {
	
	private Set<ExplorationSituation> schedule;
	private Timer t;
	private long startTime;
	private long elapsed;
	private Vavetam sys;
	private MainFrame gui;
	private boolean running = false;
	private boolean ready = false;
	private File sFile;
	private Logger log = Logger.getLogger(AutomaticSimulator.class);
	
	/**
	 * @param system the VAVETaM system that is fed with the simulated input
	 */
	public AutomaticSimulator(Vavetam system, MainFrame gui, String scheduleFileLocation) {
		sys = system;
		this.gui = gui;
		try {
			loadSchedule(new File(scheduleFileLocation));
		} catch (IllegalStateException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/**
	 * starts the simulation
	 */
	public void go() {
			if (ready) {
				log.info("started");
//				Logger.simstart(); TODO: ersetzen
//				sys.start();
				t = new Timer("Simulation Timer");
				running = true;
				t.schedule(new MyTimerTask(), 0, 20);
				startTime = System.currentTimeMillis();
		} else {
			//TODO: nicht bereit meldung zurückgeben
		}
	}
	
	/**
	 * stops the simulation
	 */
	public void stop() {
		sys.stop();
//		sys.shutDown();
//		Logger.simstop(); TODO: ersetzen
		log.trace("stopped");
		t.cancel();
		running = false;
		ready = false;
	}

	/**
	 * this subclass manages the timing of the simulated events
	 * 
	 * @author Ole Eichhorn
	 *
	 */
	private class MyTimerTask extends TimerTask {
		
		@Override
		public void run() {
			elapsed = System.currentTimeMillis() - startTime;
			gui.updateElapsedTime(elapsed);
			for (ExplorationSituation e: schedule) {
				if (e.getMilliseconds() < elapsed && !e.isDone()) {
					gui.showSituation(e);
					log.info("simulating " + e);
					if (e.isNoMEP()) {
						sys.MapExploratoryProcedureOver();
					} else {
						sys.fireMapExploratoryProcedure(e.getType(), e.getPrimaryFocusID(), Arrays.asList(e.getSecondaryFocusIDs()));
					}
					e.setDone();
					return;
				}
			}
			boolean allDone = true;
			for (ExplorationSituation e: schedule) {
				if (!e.isDone()) {
					allDone = false;
				}
			}
			if (allDone) {
				log.info("reached end of schedule");
				stop();
				sys.propagateFinishAndStop();
			}
		}
	}
	
	/**
	 * @return true, if the simulator is running
	 */
	public boolean isRunning() {
		return running;
	}
	
	/**
	 * @return true, if the simulator is ready to be started
	 */
	public boolean isReady() {
		return ready;
	}
	
	/**
	 * @param scheduleFile the schedule file to be loaded
	 * @throws IOException if the schdule file can not be loaded properly
	 * @throws IllegalStateException if a schedule is being loaded while the simulator is running
	 */
	public void loadSchedule(File scheduleFile) throws IOException, IllegalStateException {
		if (!running) {
			schedule = (new CSVParser()).loadCSVschedule(scheduleFile);
			sFile = scheduleFile;
			ready = true;
		} else {
			throw new IllegalStateException("running");
		}
	}
	
	/**
	 * resets the simulator from any state back to the start.
	 */
	public void reset() {
		if (running) {
			stop();
		}
		if (sFile != null) {
			try {
				loadSchedule(sFile);
			} catch (Exception e) {}
		}
		gui.setbackSimDisplay();
	}
}
