package vavetam.oldcode;

import static java.text.MessageFormat.format;
import static vavetam.implementation.Messages.getString;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;

import vavetam.api.structure.IDescriptor;
import vavetam.api.structure.IPredicate;
import vavetam.api.structure.IReferentialObject;



/**
 * Instances of this class are able to convert a Set of RefOs from the VAVETaM
 * system into a TeX code that can be used to show their structure in a TeX
 * document.
 * 
 * @author Ole Eichhorn
 *
 */
public class TeXCoder {

	private int counter;
	private MyFileWriter f;
	
	/**
	 * 
	 * @param mfw determines the file that will be written
	 */
	public TeXCoder(MyFileWriter mfw) {
		f = mfw;
	}
	
	/**
	 * call this method to create the TeX code and write it into a previously
	 * defined file.
	 * 
	 * @param in the Set of RefOs that is to be illustrated
	 * @param name the name of the File used to create the RefOs (mentioned as
	 * 			a caption underneath the resulting TeX graphic)
	 * @throws IOException
	 * @throws IllegalStateException
	 */
	public void createTeX(Set<IReferentialObject> in, String name) throws IOException, IllegalStateException {
		HashMap<String, String> refOsects = new HashMap<String, String>();
		f.add(getString("TeXCoder_header")); //$NON-NLS-1$
		for (IReferentialObject r: in) {
			String refOsect = ""; //$NON-NLS-1$
			int cmax = 0;
			counter = 0;
			refOsect += getString("TeXCoder_separator"); //$NON-NLS-1$
			refOsect += format(getString("TeXCoder_01"), r.getIdentifier(), att(r), counter, r.getDefiniteIdentification().replace(" ", "\\_")); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			cmax = Math.max(cmax, counter);
			counter = 1;
			refOsect += descr(r);
			cmax = Math.max(cmax, counter);
			refOsect += format(getString("TeXCoder_02"), cmax); //$NON-NLS-1$
			refOsects.put(r.getIdentifier(), refOsect);
		}
		LinkedList<String> kl = new LinkedList<String>();
		for (String s: refOsects.keySet()) {
			kl.add(s);
		}
		Collections.sort(kl);
		for (String s: kl) {
			f.add(refOsects.get(s));
		}
		f.add(getString("TeXCoder_separator")+format(getString("TeXCoder_end"), name)); //$NON-NLS-1$ //$NON-NLS-2$
		f.closeFile();
	}
	
	/**
	 * translates enums into Strings suitable for writing.
	 * the definiteness of the Designator d is returned as a String.
	 * @param d 
	 * @return
	 */
	private String def(IDescriptor d) {
		if (d != null) {
			switch (d.getDefiniteness()) {
			case ETA: return getString("TeXCoder_eta"); //$NON-NLS-1$
			case IOTA: return getString("TeXCoder_iota"); //$NON-NLS-1$
			case ALL_T: return getString("TeXCoder_allt"); //$NON-NLS-1$
			case SOME_T: return getString("TeXCoder_somet"); //$NON-NLS-1$
			default: return "";			 //$NON-NLS-1$
			}
		}
		return getString("TeXCoder_iota"); //$NON-NLS-1$
	}
	
	/**
	 * converts a Designator into a String representation
	 * 
	 * @param d
	 * @param r
	 * @return
	 */
	private String meaning(IDescriptor d, IReferentialObject r) {
		ArrayList<String> predicates = new ArrayList<String>();
		for (IPredicate p: d) {
			String name = p.getName().toString().toLowerCase().replace("_", "\\_"); //$NON-NLS-1$ //$NON-NLS-2$
			ArrayList<String> args = new ArrayList<String>();
			for (IReferentialObject r1:p.getArguments()) {
				args.add(r.equals(r1) ? "x" : format(getString("TeXCoder_09"), r1.getIdentifier())); //$NON-NLS-1$ //$NON-NLS-2$
			}
			String argsnames = StringUtils.join(args, ", "); //$NON-NLS-1$
			predicates.add(format(getString("TeXCoder_10"), name, argsnames)); //$NON-NLS-1$
		}
		String result = format((predicates.size() > 1 ? "({0})":"{0}"),  StringUtils.join(predicates, ", ")); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		return result;
	}
	
	/**
	 * creates a String representation of the attributes of a given RefO
	 * 
	 * @param r
	 * @return
	 */
	private String att(IReferentialObject r) {
		counter = 0;
		String type = r.getType().toString().toLowerCase().replace("_", "\\_"); //$NON-NLS-1$ //$NON-NLS-2$
		String result = format(getString("TeXCoder_03"), counter, type); //$NON-NLS-1$
		counter++;
		result += format(getString("TeXCoder_04"), counter); //$NON-NLS-1$
		counter++;
		if (r.getChildren().size() > 0) {
			LinkedList<String> parts = new LinkedList<String>();
			for (IReferentialObject r1:r.getChildren()) {
				parts.add(r1.getIdentifier());
			}
			Collections.sort(parts);
			String partnames = StringUtils.join(parts, ", "); //$NON-NLS-1$
			result += format(getString("TeXCoder_05"), counter, partnames); //$NON-NLS-1$
			counter++;
		}
		if (r.getParents().size() > 0) {
			for (IReferentialObject r1: r.getParents()) {
				result += format(getString("TeXCoder_06"), counter, r1.getIdentifier()); //$NON-NLS-1$
				counter++;
			}
		}
		return result+"\n"; //$NON-NLS-1$
	}
	
	private String descr(IReferentialObject r) {
		LinkedList<String> desis = new LinkedList<String>();
		String result = ""; //$NON-NLS-1$
		for (IDescriptor d: r.getDescriptors()) {
			desis.add(format(getString("TeXCoder_07"), def(d), meaning(d,r))); //$NON-NLS-1$
		}
		Collections.sort(desis);
		for (String s: desis) {
			result += format(getString("TeXCoder_08"), counter, s); //$NON-NLS-1$
			counter++;
		}
		return result;
	}
}
